#include "log-directory.h"
#include "lasterr-exception.h"

LogDirectory::LogDirectory(
   const char * init_dirname,
   const char * init_suffix,
   DWORD init_log_file_count)
 : dirname(init_dirname), suffix(init_suffix),
   log_file_count(init_log_file_count), curlog_name(NULL), curlog(NULL) {

   return;
}

LogDirectory::~LogDirectory() {
   if (curlog != NULL) {
      fclose(curlog);
   }
   delete [] curlog_name;
}

void
LogDirectory::LogException(std::exception & ee) {
   Lock();
   if (curlog != NULL) {
      const char * msg = ee.what();
      if (msg != NULL) {
         InternLogString(msg);
      } else {
         InternLogString("unidentified exception");
      }
   }
   UnLock();
}

void
LogDirectory::LogString(const char * str) {
   Lock();
   if (curlog != NULL) {
      InternLogString(str);
   }
   UnLock();
}

void
LogDirectory::LogPrintf(const char * fmt, ...) {
   Lock();
   if (curlog != NULL) {
      char buf[1024];
      va_list args;
      va_start(args, fmt);
      wvsprintf(buf, fmt, args);
      va_end(args);
      InternLogString(buf);
   }
   UnLock();
}

bool
LogDirectory::ValidLog() {
   return true;
}

void
LogDirectory::InternLogString(const char * str) {
   char buf[1024];

   wsprintf(buf, "%04d/%02d/%02d %02d:%02d:%02d UTC [%d/%d]: ",
      now.wYear, now.wMonth, now.wDay, now.wHour, now.wMinute, now.wSecond,
         GetCurrentProcessId(), GetCurrentThreadId());
   fputs(buf, curlog);
   fputs(str, curlog);
   if (str[strlen(str)-1] != '\n') {
      fputs("\n", curlog);
   }
}

void
LogDirectory::Lock() {
   EnterCriticalSection(log_lock);
   GetSystemTime(&now);
   char yymmdd[7];
   sprintf(yymmdd, "%02d%02d%02d", now.wYear % 100, now.wMonth, now.wDay);
   char * new_log_name = new char [strlen(dirname) +    // directory
                                   1 +                  // backslash
                                   6 +                  // yymmdd
                                   1 +                  // dot
                                   strlen(suffix) +     // suffix
                                   1];                  // null terminator
   strcpy(new_log_name, dirname);
   strcat(new_log_name, "\\");
   strcat(new_log_name, yymmdd);
   strcat(new_log_name, ".");
   strcat(new_log_name, suffix);
   if ((curlog == NULL) ||
       (curlog_name == NULL) ||
       (strcmp(curlog_name, new_log_name) != 0)) {
      delete [] curlog_name;
      curlog_name = new_log_name;
      if (curlog != NULL) {
         fclose(curlog);
      }
      curlog = fopen(curlog_name, "a+");
      if (log_file_count >= 2) {
         DeleteOldLogFiles();
      }
   } else {
      delete [] new_log_name;
   }
}

void
LogDirectory::UnLock() {
   if (curlog != NULL) {

      // For poor win9X users, we need to close the file, not just flush it
      // if they want to be able to see recent log entries while the service
      // is still running.

      fclose(curlog);
      curlog = NULL;
   }
   LeaveCriticalSection(log_lock);
}

struct OneLogFile {
   FILETIME ftLastWriteTime;
   char logname[7];
};

void
LogDirectory::DeleteOldLogFiles() {
   char * log_dir_pat = new char [strlen(dirname) +    // directory
                                  1 +                  // backslash
                                  2 +                  // *.
                                  strlen(suffix) +     // suffix
                                  1];                  // null terminator
   char * killname = new char [strlen(dirname) +    // directory
                               1 +                  // backslash
                               6 +                  // basename of file
                               1 +                  // dot
                               strlen(suffix) +     // suffix
                               1];                  // null terminator
   strcpy(log_dir_pat, dirname);
   strcat(log_dir_pat, "\\*.");
   strcat(log_dir_pat, suffix);
   OneLogFile ** keepers = new OneLogFile * [log_file_count];
   int curcount = 0;
   WIN32_FIND_DATA fileinfo;
   int expected_len = 7 + strlen(suffix);
   HANDLE fh = FindFirstFile(log_dir_pat, &fileinfo);
   if (fh != INVALID_HANDLE_VALUE) {
      do {
         if (strlen(fileinfo.cFileName) == expected_len) {

            // We found a new log file, make a OneLogFile struct to hold
            // the last write time and the basename of the file.

            OneLogFile * ofl = new OneLogFile;
            OneLogFile * killit = NULL;
            ofl->ftLastWriteTime = fileinfo.ftLastWriteTime;
            strncpy(ofl->logname, fileinfo.cFileName, 6);
            ofl->logname[6] = '\0';

            // The "keepers" array is a sorted list of OneLogFile pointers
            // with the oldest files at the end of the list. Do a simple
            // insertion sort to find out where this new file goes in the
            // list and insert it at this position. Once the list gets
            // log_file_count entries, the entry that would go past the end
            // of the list is deleted.

            int i;
            for (i = curcount; i >= 1; --i) {

               // If the last write time of "ofl" is less than the last
               // write time of keepers[i-1], then "ofl" is older, and we have
               // found the position it belongs in the array (i).

               if (CompareFileTime(&ofl->ftLastWriteTime,
                                   &keepers[i-1]->ftLastWriteTime) < 0) {
                  break;
               }
            }
            if (curcount == (int)log_file_count) {

               // We already have our quota of files, we have to delete
               // the oldest one...

               if (i == curcount) {

                  // The file we just looked at appears to be the oldest,
                  // so don't put it in the list, just move it directly
                  // to the killit pointer.

                  killit = ofl;
                  ofl = NULL;
               } else {

                  // The last entry in the list is the oldest file we
                  // have found so far, so kill it to make room for "ofl".

                  --curcount;
                  killit = keepers[curcount];
               }
            }
            if (ofl != NULL) {

               // Insert ofl at position 'i' in array, grow array by 1 and
               // move things down to make room.

               for (int j = curcount; j > i; --j) {
                  keepers[j] = keepers[j - 1];
               }
               ++curcount;
               keepers[i] = ofl;
            }
            if (killit != NULL) {
               strcpy(killname, dirname);
               strcat(killname, "\\");
               strcat(killname, killit->logname);
               strcat(killname, ".");
               strcat(killname, suffix);
               char buf[1024];
               wsprintf(buf, "Removing log file: %s", killname);
               InternLogString(buf);
               if (! DeleteFile(killname)) {
                  LastErrorException delerr("DeleteFile");
                  InternLogString(delerr.what());
               }
            }
         }
      } while (FindNextFile(fh, &fileinfo));
      FindClose(fh);
   }
   for (int i = 0; i < curcount; ++i) {
      delete keepers[i];
   }
   delete [] keepers;
   delete [] log_dir_pat;
   delete [] killname;
}
