
// Copyright (C) Tom Horsley, 1997. All rights reserved.

#include <windows.h>
#include <ras.h>
#include <exception>
#include "ntp-service.h"
#include "wrap-winsock.h"
#include "enable-set-time.h"
#include "raise-priority.h"
#include "ntp-packet.h"
#include "log-directory.h"
#include "wrap-socket.h"
#include "regstuff.h"
#include "lasterr-exception.h"
#include "simple-exception.h"
#include "ntp-events.h"
#include "ntp-real-events.h"
#include "kiss-exception.h"

char * NtpService::ntpversion = "4.3";

NtpService::NtpService(HINSTANCE init_hInstance)
: WrapService(init_hInstance,
              "NTPTime",
              SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_SHUTDOWN),
  ras_dll_loaded(false), ras_dll_handle(NULL),
  log_file_count(0), use_ras_dependency(true),
  retry_init_milliseconds(1000), retry_milliseconds(1000), retry_add(0),
  retry_mult(2), retry_count(0), retry_max_count(10), retry_div(1),
  log_directory_name(NULL), write_log(false), last_write_log(false),
  last_log_directory_name(NULL), ntp_server_names(NULL), ntp_server_count(0),
  ntp_server_port(123), max_adjust_seconds(0), which_server(0),
  retry_max_milliseconds(3*60*60*1000), ntp_protocol_version(3),
  allow_poke(true), allow_monitor(true), startup_delay_milliseconds(0),
  shutdown_delay_milliseconds(0), ras_delay_milliseconds(0),
  ntp_reply_port(0), disable_packet_error_checks(false),
  ntp_events(new NtpEvents), kissed_off(false)
{
   stop_event = CreateEvent(NULL, FALSE, FALSE, NULL);
}

NtpService::~NtpService()
{
   CloseHandle(stop_event);
   delete ntp_events;
}

void
NtpService::Run()
{

   // Lock the run_local critical section while we are in Run()

   WrapCriticalSection wcs(run_lock);
   ntp_events->SetInitState();
   if (startup_delay_milliseconds > 0) {
      if (WaitForSomething(startup_delay_milliseconds) == STOP_EVENT) {
         shutting_down = true;
      }
   }
   while (! shutting_down) {
      try {
         while (! shutting_down) {
            if (wait_for_ras) {
               bool loop_done = false;
               while (! IsRasActive()) {
                  switch(WaitForSomething(ras_poll_milliseconds)) {
                  case STOP_EVENT:
                     loop_done = true;
                     shutting_down = true;
                     break;
                  case POKE_EVENT:
                     loop_done = true;
                     break;
                  case TIMEOUT_EVENT:
                     break;
                  }
                  if (loop_done) {
                     break;
                  }
               }
            }
            if (shutting_down) {
               break;
            }
            if ((wait_for_ras) && (ras_delay_milliseconds > 0)) {
               if (WaitForSomething(ras_delay_milliseconds) == STOP_EVENT) {
                  shutting_down = true;
                  break;
               }
            }
            TryOnce();
            ntp_events->SetSyncOK();
            if (! shutting_down) {
               if (WaitForSomething(ntp_poll_milliseconds) == STOP_EVENT) {
                  shutting_down = true;
                  break;
               }
            } else {
               break;
            }
         }
      } catch (std::exception * ee) {

         // Better reset this in case someone poked us, but we get an
         // exception before we get a chance to call TryOnce() - wind up in
         // a pretty tight exception loop then :-).

         ntp_events->ResetPoke();
         ntp_events->SetSyncWarn();

         LogStuff * log = GetLog();
         log->LogException(*ee);
         delete ee;
         if (kissed_off) {

            // If the server we just talked to kissed us off, remove
            // him from the server list.

            kissed_off = false;
            for (int i = which_server + 1; i < ntp_server_count; ++i) {
               ntp_server_names[i - 1] = ntp_server_names[i];
            }
            ntp_server_count -= 1;
            if (which_server == 0) {
               which_server = ntp_server_count - 1;
            } else {
               which_server -= 1;
            }
            if (ntp_server_count <= 0) {
               shutting_down = true;
               ntp_events->SetShutdownState();
               log->LogString("all servers sent kiss-o-death, shutting down");
            }
         }
         if (! shutting_down) {

            // When we get an error, want to just retry up to
            // retry_max_count times (or forever if retry_max_count is
            // zero). Adjust the next retry time based on the retry_add,
            // mul, and div parameters.

            ++retry_count;
            if ((retry_max_count != 0) && (retry_count >= retry_max_count)) {
               shutting_down = true;
               ntp_events->SetShutdownState();
               log->LogString("shutting down NTP service");
            } else {

               // Try every server once before bumping up retry time

               ++which_server;
               DWORD this_retry_ms = retry_milliseconds;
               if (which_server >= ntp_server_count) {
                  which_server = 0;
                  retry_milliseconds =
                     ((retry_milliseconds + retry_add) * retry_mult) /
                        retry_div;
                  if (retry_milliseconds > retry_max_milliseconds) {
                     retry_milliseconds = retry_max_milliseconds;
                  }
                  if (retry_milliseconds < retry_init_milliseconds) {
                     retry_milliseconds = retry_init_milliseconds;
                  }
               }
               if (WaitForSomething(this_retry_ms) == STOP_EVENT) {
                  shutting_down = true;
                  log->LogString("shutting down NTP service");
                  break;
               }
            }
         }
      }
   }
}

bool
NtpService::OnInit()
{
   bool rval = true;
   try {
      shutting_down = false;
      RegistryLocalMachine lm;
      RegistryKey ntproot(lm, "Software\\ntcli\\NTPTime");
      RegistryKey ntpkey(ntproot, ntpversion);
      DWORD temp;
      temp = ntpkey.GetRegDwordValue("WriteLogEntries");
      if (temp != 0) {
         log_directory_name = ntpkey.GetRegExpandSzValue("LogDirectory");
      }
      write_log = (temp != 0);
      temp = ntpkey.GetRegDwordValue("WaitForRAS");
      wait_for_ras = (temp != 0);
      if (wait_for_ras) {
         ras_poll_milliseconds =
            ntpkey.GetRegDwordValue("RASPollMilliseconds");
      }
      temp = ntpkey.GetRegDwordValue("RunOnce");
      run_once = (temp != 0);

      // Pick multiple space, comma, or semicolon names out of server name
      // string.

      char * ntp_server_str = ntpkey.GetRegSzValue("NTPServerName");
      int max_count = 1;
      char c;
      for (char * s = ntp_server_str; (c = *s) != '\0'; ++s) {
         if ((c == ' ') || (c == ';') || (c == ',')) {
            ++max_count;
         }
      }
      ntp_server_names = new char * [max_count];
      bool skipping = true;
      for ( ; ; ) {
         if (skipping) {
            while (((c = *ntp_server_str) == ' ') || (c == ',') || (c == ';')) {
               *ntp_server_str++ = '\0';
            }
            if (c == '\0') {
               break;
            } else {
               if (ntp_server_count >= max_count) {
                  throw new
                     SimpleException("Internal error parsing server names");
               }
               ntp_server_names[ntp_server_count++] = ntp_server_str;
               skipping = false;
            }
         } else {
            while (! (((c = *ntp_server_str) == ' ') || (c == ',') ||
                      (c == ';') || (c == '\0'))) {
               ++ntp_server_str;
            }
            if (c == '\0') {
               break;
            } else {
               *ntp_server_str++ = '\0';
               skipping = true;
            }
         }
      }
      if (ntp_server_count <= 0) {
         throw new SimpleException("Unable to parse any server names");
      }
      ntp_poll_milliseconds = ntpkey.GetRegDwordValue("NTPPollMilliseconds");
      DWORD fmt;
      DWORD len;
      if ((ntpkey.QueryValue("NTPServerPort", fmt, len)) &&
          (fmt == REG_DWORD)) {
         ntp_server_port = (unsigned short)
            ntpkey.GetRegDwordValue("NTPServerPort");
      }
      if ((ntpkey.QueryValue("MaxAdjustSeconds", fmt, len)) &&
          (fmt == REG_DWORD)) {
         max_adjust_seconds = ntpkey.GetRegDwordValue("MaxAdjustSeconds");
      }
      if ((ntpkey.QueryValue("LogFileCount", fmt, len)) &&
          (fmt == REG_DWORD)) {
         log_file_count = ntpkey.GetRegDwordValue("LogFileCount");
      }
      if ((ntpkey.QueryValue("RetryMinMilliseconds", fmt, len)) &&
          (fmt == REG_DWORD)) {
         retry_init_milliseconds =
            ntpkey.GetRegDwordValue("RetryMinMilliseconds");
      }
      if ((ntpkey.QueryValue("RetryMaxMilliseconds", fmt, len)) &&
          (fmt == REG_DWORD)) {
         retry_max_milliseconds =
            ntpkey.GetRegDwordValue("RetryMaxMilliseconds");
      }
      if ((ntpkey.QueryValue("RetryAdd", fmt, len)) && (fmt == REG_DWORD)) {
         retry_add = ntpkey.GetRegDwordValue("RetryAdd");
      }
      if ((ntpkey.QueryValue("RetryMult", fmt, len)) && (fmt == REG_DWORD)) {
         retry_mult = ntpkey.GetRegDwordValue("RetryMult");
      }
      if ((ntpkey.QueryValue("RetryDiv", fmt, len)) && (fmt == REG_DWORD)) {
         retry_div = ntpkey.GetRegDwordValue("RetryDiv");
      }
      if ((ntpkey.QueryValue("RetryMaxCount", fmt, len)) &&
          (fmt == REG_DWORD)) {
         retry_max_count = ntpkey.GetRegDwordValue("RetryMaxCount");
      }
      if ((ntpkey.QueryValue("NtpProtocolVersion", fmt, len)) &&
          (fmt == REG_DWORD)) {
         ntp_protocol_version =
            (unsigned char)ntpkey.GetRegDwordValue("NtpProtocolVersion");
      }
      if ((ntpkey.QueryValue("AllowPoke", fmt, len)) &&
          (fmt == REG_DWORD)) {
         allow_poke = (ntpkey.GetRegDwordValue("AllowPoke") != 0);
      }
      if ((ntpkey.QueryValue("AllowMonitor", fmt, len)) &&
          (fmt == REG_DWORD)) {
         allow_monitor = (ntpkey.GetRegDwordValue("AllowMonitor") != 0);
      }
      if ((ntpkey.QueryValue("StartupDelayMilliseconds", fmt, len)) &&
          (fmt == REG_DWORD)) {
         startup_delay_milliseconds =
            ntpkey.GetRegDwordValue("StartupDelayMilliseconds");
      }
      if ((ntpkey.QueryValue("RasDelayMilliseconds", fmt, len)) &&
          (fmt == REG_DWORD)) {
         ras_delay_milliseconds =
            ntpkey.GetRegDwordValue("RasDelayMilliseconds");
      }
      if ((ntpkey.QueryValue("ShutdownDelayMilliseconds", fmt, len)) &&
          (fmt == REG_DWORD)) {
         shutdown_delay_milliseconds =
            ntpkey.GetRegDwordValue("ShutdownDelayMilliseconds");
      }
      if ((ntpkey.QueryValue("NtpReplyPort", fmt, len)) &&
          (fmt == REG_DWORD)) {
         ntp_reply_port = ntpkey.GetRegDwordValue("NtpReplyPort");
      }
      if ((ntpkey.QueryValue("DisablePacketErrorChecks", fmt, len)) &&
          (fmt == REG_DWORD)) {
         disable_packet_error_checks =
            (ntpkey.GetRegDwordValue("DisablePacketErrorChecks") != 0);
      }
      GetLog()->LogPrintf("Starting up NTP service version %s", ntpversion);
      for (int i = 0; i < ntp_server_count; ++i) {
         if (i == 0) {
            GetLog()->LogPrintf("    Primary server: %s", ntp_server_names[i]);
         } else {
            GetLog()->LogPrintf("   Backup server %d: %s",
                                i, ntp_server_names[i]);
         }
      }
      if (allow_poke || allow_monitor) {

         // If we are going to allow event interaction via ntptray, replace
         // the stub NtpEvents object with one that really creates the event
         // handles to the named events.

         NtpRealEvents * new_events =
            new NtpRealEvents(IsRunningOnNT(), allow_poke, allow_monitor,
                              GetLog());
         if (new_events->IsValid()) {
            delete ntp_events;
            ntp_events = new_events;
         } else {

            // If we weren't able to create the valid events with proper
            // security descriptors, then just pretend AllowPoke was false
            // and try to keep running.

            delete new_events;
            GetLog()->LogPrintf("Unable to create valid NtpRealEvents, "
                                "ntpoke and ntptray disabled.");
         }
      }
   } catch (std::exception * ee) {
      rval = false;
      GetLog()->LogException(*ee);
      delete ee;
   }
   return rval;
}

void
NtpService::OnStop()
{
   shutting_down = true;
   SetEvent(stop_event);
   {
      // Wait for run_lock to make sure the Run() routine has finished
      // before we tell the service manager we actually stopped.

      WrapCriticalSection wcs(run_lock);
   }
   ntp_events->SetShutdownState();
   GetLog()->LogString("NTP stop request");
}

void
NtpService::OnInterrogate()
{
}

void
NtpService::OnPause()
{
}

void
NtpService::OnContinue()
{
}

void
NtpService::OnShutdown()
{
   shutting_down = true;
   SetEvent(stop_event);
   {
      // Wait for the run_lock critical section to make sure the Run()
      // routine has finished before we tell the service manager we shut
      // down.

      WrapCriticalSection wcs(run_lock);
   }
   ntp_events->SetShutdownState();
   GetLog()->LogString("NTP shutdown request");
}

bool
NtpService::OnUserControl(DWORD user_opcode)
{
   return true;
}

void
NtpService::TryOnce()
{
   // We leave the poke_event set (if it was set) until we try
   // to get the time, then reset it here so WaitForSomething
   // will start waiting again

   kissed_off = false;
   ntp_events->ResetPoke();

   SYSTEMTIME right_now;
   WrapWinsock ww;   // Initialize winsock services...

   struct sockaddr_in serv_addr;
   memset((void *)&serv_addr, 0, sizeof(serv_addr));
   serv_addr.sin_family = AF_INET;
   serv_addr.sin_port = htons(ntp_server_port);
   char * ntp_server_name = ntp_server_names[which_server];
   GetLog()->LogPrintf("Requesting time from server: %s", ntp_server_name);
   struct hostent * hp = gethostbyname(ntp_server_name);
   if (hp == NULL) {
      DWORD errcode = GetLastError();

      // Don't barf just yet. Maybe this is an absolute IP address string...

      unsigned long ipaddr = inet_addr(ntp_server_name);
      if (ipaddr != INADDR_NONE) {
         memcpy((void *)&serv_addr.sin_addr.s_addr, (void *)&ipaddr,
                sizeof(serv_addr.sin_addr.s_addr));
      } else {
         throw new LastErrorException(errcode, "gethostbyname(%s)",
                                      ntp_server_name);
      }
   } else {
      memcpy((void *)&serv_addr.sin_addr.s_addr, (void *)hp->h_addr,
             sizeof(serv_addr.sin_addr.s_addr));
   }

   WrapSocket s(socket(AF_INET, SOCK_DGRAM, 0));
   if ((SOCKET)s == INVALID_SOCKET) {
      throw new LastErrorException("socket");
   }

   struct sockaddr_in addr;
   memset((char *)&addr, 0, sizeof addr);
   addr.sin_family = AF_INET;
   addr.sin_port = htons((short)ntp_reply_port);
   addr.sin_addr.s_addr = htonl(INADDR_ANY);
   if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) == SOCKET_ERROR) {
      throw new LastErrorException("bind");
   }

   int on = 1;
   if (ioctlsocket(s, FIONBIO, (u_long *) &on) == SOCKET_ERROR) {
      throw new LastErrorException("ioctlsocket");
   }

   struct timeval ten;
   ten.tv_sec = 10;
   ten.tv_usec = 0;
   fd_set reader;
   FD_ZERO(&reader);
   FD_SET(s, &reader);

   RaisePriority rp; // raise process priority to max while diddling time

   BasicNtpPacket cpkt;
   FullNtpPacket  spkt;
   memset((void *)&cpkt, 0, sizeof(cpkt));
   SetPacketLeapVersionMode(cpkt,0,ntp_protocol_version,3);
   GetSystemTime(&right_now);
   NtpLongFixed rn(right_now);
   rn.ConvertToNetworkOrder(cpkt.transmit_timestamp);

   if (sendto(s, (char *)&cpkt, sizeof(cpkt), 0,
        (struct sockaddr *)&serv_addr, sizeof(serv_addr)) == SOCKET_ERROR) {
      throw new LastErrorException("sendto");
   }

   int select_status = select(1, &reader, NULL, NULL, &ten);
   if (select_status == SOCKET_ERROR) {
      throw new LastErrorException("select");
   }
   SYSTEMTIME recvtime;
   GetSystemTime(&recvtime);
   if (select_status == 0) {

      // I seem to have timed out without getting a return packet. Just
      // log it and get outta here

      throw new SimpleException("no response from NTP server");
   }
   struct sockaddr_in gotfrom;
   int gotlen = sizeof(gotfrom);
   if (recvfrom(s, (char *)&spkt, sizeof(spkt), 0,
          (struct sockaddr *)&gotfrom, &gotlen) == SOCKET_ERROR) {
      throw new LastErrorException("recvfrom");
   }

   // The draft rfc-xxxx revision for the sntp standard says that servers
   // can send us a Kiss 'o Death packet by sticking zero in the stratum and
   // a "kiss code" in the reference_identifier field. We should stop
   // pestering the server if it kisses us off.

   if ((spkt.basic_part.stratum == 0) &&
       ((spkt.basic_part.reference_identifier[0] != 0) ||
        (spkt.basic_part.reference_identifier[1] != 0) ||
        (spkt.basic_part.reference_identifier[2] != 0) ||
        (spkt.basic_part.reference_identifier[3] != 0))) {
      kissed_off = true;
      throw new KissODeathException(ntp_server_name,
                                    spkt.basic_part.reference_identifier);
   }

   // Need to do minimal validation of incoming NTP packet.
   // Do basic checking on returned value to see if it is valid per
   // description in RFC1769...

   if (! disable_packet_error_checks) {
      if (PacketLeapIndicator(spkt.basic_part) == LI_ALARM) {
         throw new SimpleException("LI_ALARM set (server not synchronized)");
      }
      if ((spkt.basic_part.stratum < 1) || (spkt.basic_part.stratum > 15)) {
         throw new SimpleException("invalid server stratum");
      }
   }

   NtpLongFixed T1(spkt.basic_part.originate_timestamp);
   NtpLongFixed T2(spkt.basic_part.receive_timestamp);
   NtpLongFixed T3(spkt.basic_part.transmit_timestamp);
   NtpLongFixed T4(recvtime);

   // And one more validity check --

   if ((T3.FractionalPart() == 0) && (T3.IntegerPart() == 0)) {
      throw new SimpleException("invalid (zero) timestamp from server");
   }

   NtpLongFixed d = (T4 - T1) - (T3 - T2);
   NtpLongFixed t = ((T2 - T1) + (T3 - T4)).SignedDivByTwo();

   if (max_adjust_seconds > 0) {
      // If there is a configured max adjustment limit, check it
      NtpLongFixed m(t);
      if (m.IntegerPart() & 0x80000000) {
         m = - m;
      }
      if (m.IntegerPart() > max_adjust_seconds) {
         GetLog()->LogPrintf("time adjustment of %ld seconds > max of %ld",
                                m.IntegerPart(), max_adjust_seconds);
         throw new SimpleException("time adjustment out of range");
      }
   }

   EnableSetTime est; // Diddle privileges so I can set the time...

   GetSystemTime(&right_now);
   NtpLongFixed new_clock(right_now);
   new_clock = new_clock + t;
   new_clock.ConvertToSYSTEMTIME(right_now);
   if (! SetSystemTime(&right_now)) {
      throw new LastErrorException("SetSystemTime");
   }
   LogTimeDiff("clock adjustment", t);
   LogTimeDiff("   round trip delay", d);
   GetLog()->LogPrintf("   server stratum: %u", spkt.basic_part.stratum);
   ResetRetry();

   // I seem to have actually set the time. If I was only supposed to do this
   // once, then trigger a shutdown.

   if (run_once) {
      if (shutdown_delay_milliseconds > 0) {
         WaitForSomething(shutdown_delay_milliseconds);
      }
      shutting_down = true;
      ntp_events->SetShutdownState();
      GetLog()->LogString("Shutting down after setting time once.");
   }

   // At this return (or at any 'throw' above), all the destructors should
   // tun back off all the privileges, raised priority, winsock activation,
   // etc.

   return;
}

LogStuff *
NtpService::GetLog()
{
   // If we happen to call this before we have read the proper log file
   // information from the registry, then make sure we change to the correct
   // log directory once we log something after reading the registry
   // parameters. (This only happens when I throw in logs of log messages
   // for debugging, but it keeps me from blowing up if I accidentally
   // toss in a log entry earlier than I should).

   if ((log_stuff == NULL) || (last_write_log != write_log) ||
       (last_log_directory_name != log_directory_name)) {
      delete log_stuff;
      if (write_log && (log_directory_name != NULL)) {
         log_stuff =
            new LogDirectory(log_directory_name, "ntp", log_file_count);
      } else {
         log_stuff = new LogStuff;
      }
      last_write_log = write_log;
      last_log_directory_name = log_directory_name;
   }
   return log_stuff;
}

LPCTSTR
NtpService::GetDependencyList()
{
   // If RAS is installed on this system, then make sure we depend on it so
   // it will get started before we start trying to poll the RAS
   // configuration info. If RAS isn't installed, I guess we'll be using a
   // direct internet connection, so we won't try to call the RAS polling
   // code anyway...

   if (use_ras_dependency && IsInstalled("RasMan")) {
      return "RasMan\0\0";
   } else {
      return WrapService::GetDependencyList();
   }
}

LPTSTR
NtpService::GetDescriptiveText() {
   return "Simple Network Time Protocol client for setting system time.";
}

void
NtpService::SetUseRasDependency(bool newdep)
{
   use_ras_dependency = newdep;
}

void
NtpService::LogTimeDiff(const char * msg, NtpLongFixed & t)
{
   int neg = 0;
   NtpLongFixed x(t);
   if (t.IntegerPart() & 0x80000000) {
      neg = 1;
      x = - x;
   }
   SYSTEMTIME st;
   NtpLongFixed fx(0, x.FractionalPart());
   fx.ConvertToSYSTEMTIME(st);
   GetLog()->LogPrintf("%s: %s%ld.%03d seconds",
                          msg, (neg?"-":""), x.IntegerPart(),
                             st.wMilliseconds);
}

void
NtpService::ResetRetry()
{
   // This routine is called when we set the time successfully.
   // It resets all the retry parameters.

   retry_milliseconds = retry_init_milliseconds;
   retry_count = 0;
   which_server = 0;
}

NtpService::WaitReasons
NtpService::WaitForSomething(DWORD milliseconds) {
   HANDLE ev[3];
   ev[0] = stop_event;
   DWORD evcount = ntp_events->AppendExtraEvents(ev, 1);
   DWORD rval = WaitForMultipleObjects(evcount, ev, FALSE, milliseconds);
   if ((rval == WAIT_OBJECT_0) || (rval == WAIT_ABANDONED_0)) {
      return STOP_EVENT;
   }
   if ((evcount >= 2) &&
       (rval == (WAIT_OBJECT_0 + 1)) || (rval == (WAIT_ABANDONED_0 + 1))) {
      return POKE_EVENT;
   }
   if (rval == WAIT_TIMEOUT) {
      return TIMEOUT_EVENT;
   }
   throw new LastErrorException("WaitForMultipleObjects");
}
