
// Portions Copyright (C) Tom Horsley, 1997. All rights reserved.
// Other portions adapted from the C++ server class sample code from the
// MSDN library.

#include "wrap-service.h"
#include "wrap-sc-handle.h"
#include "lasterr-exception.h"
#include "log-stuff.h"

WrapService* WrapService::this_service = NULL;

bool WrapService::w95_posted_quit = false;

WrapService::WrapService(
   HINSTANCE init_hInstance,
   const char* init_service_name,
   DWORD init_controls_accepted)
: service_name(init_service_name),
  service_status(NULL),
  is_running_on_nt(false),
  run_on_console(false),
  debug_level(0),
  already_stopped(false),
  hInstance(init_hInstance),
  has_error(false),
  log_stuff(NULL),
  w95_window(NULL) {
   OSVERSIONINFO osvi;
   osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
   if (GetVersionEx(&osvi)) {
      is_running_on_nt = (osvi.dwPlatformId == VER_PLATFORM_WIN32_NT);
   }

   this_service = this;

   status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
   status.dwCurrentState = SERVICE_STOPPED;
   status.dwControlsAccepted = init_controls_accepted;
   status.dwWin32ExitCode = 0;
   status.dwServiceSpecificExitCode = 0;
   status.dwCheckPoint = 0;
   status.dwWaitHint = 0;
}

WrapService::~WrapService() {
   delete log_stuff;
   return;
}

bool
WrapService::IsInstalled(const char * alt_name) {
   if (! IsRunningOnNT()) {
      // On 95, if you can run the program, it must be installed :-).
      return true;
   }
   if (alt_name == NULL) {
      alt_name = service_name;
   }
   bool rval = false;
   WrapScHandle scm(OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS));
   if (scm != NULL) {
      WrapScHandle service(OpenService(scm, alt_name,
                                          SERVICE_QUERY_CONFIG));
      if (service != NULL) {
         QUERY_SERVICE_CONFIG qsc;
         DWORD dummy;
         if (QueryServiceConfig(service, &qsc, sizeof(qsc), &dummy)) {
            if (qsc.dwStartType == SERVICE_DISABLED) {
               return false;
            }
         }
         rval = true;
      }
   }
   return rval;
}

bool
WrapService::Install(DWORD start_type) {
   if (! IsRunningOnNT()) {
      // On 95, nothing magical needed to install.
      return true;
   }
   WrapScHandle scm(OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS));
   if (scm == NULL) return false;

   char file_path[_MAX_PATH];
   char * file_path_ptr = file_path;
   GetModuleFileName(NULL, file_path, _MAX_PATH);

   // Ratchin, satcha, matten, frazza... Microsoft (in its infinite wisdom)
   // was just so proud of allowing spaces in filenames (something Unix did
   // forever, but was sensible enough not to use everywhere), that they
   // decided to put spaces in just about every standard directory name they
   // use for installing software (Hey! Everybody! Look at me! Look what I
   // can do! Hey! Com'on! Look at me! Aw, com'on! Look at me guys...).

   // Having allowed spaces, they then failed to find every API that wanted
   // a path name and fix it to properly quote spaces if it needed them
   // quoted, thus leaving a mess for all the programmers in the world to
   // clean up after (and Bill Gates wonders why people hate him :-).

   // The following code fixes the mess that is CreateService()...

   int escape_count = 0;
   int dir_count = 0;
   int path_len = 0;
   char * fpp = file_path;
   char c;
   while ((c = *fpp++) != '\0') {
      ++path_len;
      if ((c == ' ') || (c == '\t') || (c == '\"')) {
         ++escape_count;
      }
      if (c == '\\') {
         ++dir_count;
      }
   }
   if (escape_count > 0) {
      file_path_ptr = new char[path_len + escape_count + dir_count + 3];
      fpp = file_path;
      char * dest = file_path_ptr;
      *dest++ = '\"';
      while ((c = *fpp++) != '\0') {

         // I don't know that you can put a " in a filename, but as long as
         // I'm using a sledge hammer anyway....

         if ((c == '\"') || (c == '\\')) {
            *dest++ = '\\';
         }
         *dest++ = c;
      }
      *dest++ = '\"';
      *dest = '\0';
   }

   // file_path_ptr is now the original file name if it didn't need quoting,
   // or a properly quoted version of the name, and it is now safe to pass
   // it to the CreateService routine.

   WrapScHandle service(CreateService(scm,
                                      service_name,
                                      service_name,
                                      SERVICE_ALL_ACCESS,
                                      SERVICE_WIN32_OWN_PROCESS,
                                      start_type,
                                      SERVICE_ERROR_NORMAL,
                                      file_path_ptr,
                                      NULL,
                                      NULL,
                                      GetDependencyList(),
                                      NULL,
                                      NULL));
   if (service == NULL) {
      return false;
   } else {
      LPTSTR desc = GetDescriptiveText();
      if (desc != NULL) {

         // Let's find out if we are running on a version of Windows
         // which has the ChangeServiceConfig2 API. If so, call it
         // to record the description.

         HINSTANCE advapi_dll = LoadLibrary("Advapi32");
         if (advapi_dll != NULL) {
            FARPROC config2_ptr = GetProcAddress(advapi_dll,
                                     "ChangeServiceConfig2A");
            if (config2_ptr != NULL) {
               SERVICE_DESCRIPTION sd;
               sd.lpDescription = desc;
               typedef BOOL WINAPI
                  ChangeServiceConfig2_func(SC_HANDLE, DWORD, LPVOID);
               ChangeServiceConfig2_func * csc2 =
                  (ChangeServiceConfig2_func *)config2_ptr;
               csc2(service, SERVICE_CONFIG_DESCRIPTION, (LPVOID)&sd);
            }
            FreeLibrary(advapi_dll);
         }
      }
   }
   return true;
}

LPCTSTR
WrapService::GetDependencyList() {
   return NULL;
}

LPTSTR
WrapService::GetDescriptiveText() {
   return NULL;
}

bool
WrapService::Uninstall() {
   if (! IsRunningOnNT()) {
      // On 95, nothing magical needed to un-install...
      return true;
   }
   WrapScHandle scm(OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS));
   if (scm == NULL) return false;

   bool rval = false;
   WrapScHandle service(OpenService(scm, service_name,
                                    DELETE|SERVICE_QUERY_STATUS));
   if (service != NULL) {

      // If the service isn't stopped, we (apparently) don't get an error
      // from the DeleteService request, but it doesn't really uninstall it,
      // so let's throw in an explicit check for service stopped before
      // trying to do the delete.

      SERVICE_STATUS ss;
      if (QueryServiceStatus(service, &ss) &&
          (ss.dwCurrentState == SERVICE_STOPPED) &&
          DeleteService(service)) {
         rval = true;
      }
   }
   return rval;
}

bool
WrapService::StartService() {
   SERVICE_TABLE_ENTRY st[] = {
      {(char *)service_name, ServiceMain},
      {NULL, NULL}
   };

   if (IsRunningOnNT()) {
      // On NT, go through normal service dispatcher rigamarole.
      return (StartServiceCtrlDispatcher(st) != 0);
   } else {

      DWORD thread_id;
      CreateThread(NULL, 0, Win95ServiceThread, (LPVOID)this, 0, &thread_id);

      WNDCLASSEX  wndclass;
      memset((void *)&wndclass, 0, sizeof(wndclass));
      wndclass.cbSize = sizeof(wndclass);
      wndclass.style = CS_HREDRAW | CS_VREDRAW;
      wndclass.lpfnWndProc = Win95WindowProc;
      wndclass.hInstance = hInstance;
      wndclass.lpszClassName = service_name;
      RegisterClassEx(&wndclass);
      w95_window = CreateWindowEx(WS_EX_TOOLWINDOW,
                                  service_name,
                                  service_name,
                                  WS_POPUP | WS_SIZEBOX | WS_SYSMENU,
                                  0, 0, 0, 0,
                                  NULL,
                                  NULL,
                                  hInstance,
                                  NULL);
      ShowWindow(w95_window, SW_HIDE);
      UpdateWindow(w95_window);
      MSG msg;

      while (GetMessage(&msg, NULL, 0, 0)) {
         TranslateMessage(&msg);
         DispatchMessage(&msg);
      }

      return true;
   }
   return false;
}

void
WrapService::SetStatus(DWORD new_status) {
   WrapCriticalSection wcs(stat_lock);
   if (! ((new_status == SERVICE_STOPPED) && already_stopped)) {
      status.dwCurrentState = new_status;
      if (IsRunningOnNT()) {
         if (status.dwCurrentState == SERVICE_STOPPED) {
            already_stopped = true;
            if (status.dwServiceSpecificExitCode == 0) {
               status.dwServiceSpecificExitCode = 1;
            }
         }
         SetServiceStatus(service_status, &status);
      }
   }
}

bool
WrapService::Initialize() {
   SetStatus(SERVICE_START_PENDING);

   bool rval = OnInit();

   status.dwWin32ExitCode = GetLastError();
   status.dwCheckPoint = 0;
   status.dwWaitHint = 0;
   if (! rval) {
      SetStatus(SERVICE_STOPPED);
      return false;
   }

   SetStatus(SERVICE_RUNNING);

   return true;
}

bool
WrapService::OnInit() {
   return true;
}

void
WrapService::OnStop() {
   return;
}

void
WrapService::OnInterrogate() {
   return;
}

void
WrapService::OnPause() {
   return;
}

void
WrapService::OnContinue() {
   return;
}

void
WrapService::OnShutdown() {
   return;
}

bool
WrapService::OnUserControl(DWORD /* user_opcode */) {
   return false;
}

LogStuff *
WrapService::GetLog() {
   if (log_stuff == NULL) {
      log_stuff = new LogStuff;
   }
   return log_stuff;
}

bool
WrapService::Debugging(int level) {
   return level <= debug_level;
}

void
WrapService::SetDebugLevel(int new_level) {
   debug_level = new_level;
}

void
WrapService::RunOnConsole() {
   run_on_console = true;
}

void WINAPI
WrapService::ServiceMain(DWORD /* argc */, LPTSTR* /* argv */) {
   this_service->status.dwCurrentState = SERVICE_START_PENDING;
   this_service->service_status =
      RegisterServiceCtrlHandler(this_service->service_name, Handler);

   if (this_service->service_status == NULL) {
      return;
   }

   if (this_service->Initialize()) {
      this_service->status.dwWin32ExitCode = 0;
      this_service->status.dwCheckPoint = 0;
      this_service->status.dwWaitHint = 0;
      this_service->Run();
   }

   this_service->SetStatus(SERVICE_STOPPED);
}

void WINAPI
WrapService::Handler(DWORD opcode) {
   switch (opcode) {
   case SERVICE_CONTROL_STOP:
      this_service->SetStatus(SERVICE_STOP_PENDING);
      this_service->OnStop();
      break;

   case SERVICE_CONTROL_PAUSE:
      this_service->OnPause();
      break;

   case SERVICE_CONTROL_CONTINUE:
      this_service->OnContinue();
      break;

   case SERVICE_CONTROL_INTERROGATE:
      this_service->OnInterrogate();
      break;

   case SERVICE_CONTROL_SHUTDOWN:
      this_service->OnShutdown();
      break;

   default:
      this_service->OnUserControl(opcode);
      break;
   }
   this_service->SetStatus(this_service->status.dwCurrentState);
}

DWORD WINAPI
WrapService::Win95ServiceThread(LPVOID arg) {
   return ((WrapService *)arg)->Win95ServiceFunc();
}

DWORD
WrapService::Win95ServiceFunc() {
   try {
      if (OnInit()) {
         Run();
         SetStatus(SERVICE_STOPPED);
      }
   } catch (std::exception * ee) {
      GetLog()->LogPrintf("Error thrown in Win95ServiceFunc: %s", ee->what());
      AppendError(ee->what());
      delete ee;
   }
   if (w95_window != NULL) {
      PostMessage(w95_window, WM_CLOSE, 0, 0);
   }
   return 0;
}

LRESULT CALLBACK
WrapService::Win95WindowProc(
   HWND hwnd,
   UINT iMsg,
   WPARAM wParam,
   LPARAM lParam) {

   if (! w95_posted_quit) {
      bool i_am_done = false;
      switch(iMsg) {
      case WM_CLOSE:
      case WM_DESTROY:
      case WM_QUERYENDSESSION:
         i_am_done = true;
         break;
      case WM_SYSCOMMAND:
         i_am_done = (wParam == SC_CLOSE);
         break;
      }
      if (i_am_done) {
         this_service->OnShutdown();
         w95_posted_quit = true;
         PostQuitMessage(0);
      }
   }
   return DefWindowProc(hwnd, iMsg, wParam, lParam);
}

void
WrapService::AppendError(const char * str) {
   WrapCriticalSection grab_lock(error_lock);
   error_reason.append(str);
   has_error = true;
}

void
WrapService::AppendError(const std::string & str) {
   WrapCriticalSection grab_lock(error_lock);
   error_reason.append(str);
   has_error = true;
}
