
// Copyright (C) Tom Horsley, 1998. All rights reserved.

#include "wincmd.h"
#include "ftrace.h"
#include "simple-exception.h"
#include <ctype.h>
#include <string.h>

Wincmd::Wincmd(PSTR szCmdLine) {
   FTBEG(("Wincmd::Wincmd(%s)", szCmdLine))
   skip = 0;
   int state = 1;
   std::string temp;
   char c;
   bool have_arg = false;

   // Run a little state machine to parse the arguments pretty much the
   // same way the C runtime does for console programs

   while ((c = *szCmdLine++) != '\0') {
      switch(state) {
      case 1:           // skipping space between args
         if (isspace(c)) {
            state = 1;
         } else {
            have_arg = 1;
            if (c == '\"') {
               state = 4;
            } else if (c == '\\') {
               state = 3;
            } else {
               state = 2;
               temp.append(1, c);
            }
         }
         break;
      case 2:           // accumulating un-quoted arg
         if (isspace(c)) {
            state = 1;
            argv.push_back(temp);
            temp = "";
            have_arg = false;
            state = 1;
         } else if (c == '\\') {
            state = 3;
         } else if (c == '\"') {
            state = 4;
         } else {
            temp.append(1, c);
            state = 2;
         }
         break;
      case 3:           // saw backslash while not in quoted string
         if (c == '\"') {
            temp.append(1, '\"');
            state = 2;
         } else {
            temp.append(1, '\\');
            --szCmdLine;
            state = 2;
         }
         break;
      case 4:           // inside quoted string
         if (c == '\"') {
            state = 2;
         } else if (c == '\\') {
            state = 5;
         } else {
            temp.append(1, c);
            state = 4;
         }
         break;
      case 5:           // saw backslash inside quoted string
         if (c == '\"') {
            temp.append(1, '\"');
            state = 4;
         } else {
            temp.append(1, '\\');
            --szCmdLine;
            state = 4;
         }
         break;
      }
   }
   if (have_arg) {
      argv.push_back(temp);
   }
   FTEND
}

Wincmd::~Wincmd() {
   FTBEG(("Wincmd::~Wincmd()"))
   return;
   FTEND
}

int
Wincmd::Argc() {
   FTBEG(("Wincmd::Argc()"))
   return argv.size() - skip;
   FTEND
}

std::string &
Wincmd::operator[](int i) {
   FTBEG(("Wincmd::operator[](%d)", i))
   return argv[skip + i];
   FTEND
}

void
Wincmd::InterpretArgs(ArgDef * argdef, int argdef_count) {
   FTBEG(("Wincmd::InterpretArgs()"))
   int i;
   for (i = 0; i < argdef_count; ++i) {
      argdef[i].arg_origin = WINCMD_UNSPECIFIED;
   }
   for ( ; ; ) {
      if (skip >= argv.size()) {
         break;
      }
      const char * curarg = argv[skip].c_str();

      // Allow args to start with either a - or a /

      if ((*curarg == '/') || (*curarg == '-')) {

         // This looks like it must be an argument name, so see if we
         // can find the matching argument in the argdef array.

         ++skip;
         ++curarg;

         int bestmatch = -1;
         bool negate = false;
         int curlen = strlen(curarg);

         for (i = 0; i < argdef_count; ++i) {

            // look for exact match first, followed by case insensitive
            // match.

            int arglen = strlen(argdef[i].arg_name);
            if (curlen >= arglen) {

               // Allow argument value to follow name with an = or :
               // separating them, or simply allow next positional argument
               // to be used as arg value.

               if ((curarg[arglen] == '=') ||
                   (curarg[arglen] == ':') ||
                   (curarg[arglen] == '\0')) {
                  if (strncmp(curarg, argdef[i].arg_name, arglen) == 0) {
                     bestmatch = i;
                     negate = false;
                     break;
                  } else if (_strnicmp(curarg, argdef[i].arg_name, arglen) ==
                                0) {
                     bestmatch = i;
                     negate = false;
                  }
               }
               if (bestmatch == -1) {
                  if ((argdef[i].arg_kind == WINCMD_BOOL) &&
                      (_strnicmp(curarg, "no", 2) == 0)) {

                     // Allow boolean argument to be negated by giving
                     // the arg name with a no no- or no_ prefix

                     const char * noarg = curarg + 2;
                     if ((*noarg == '-') || (*noarg == '_')) {
                        ++noarg;
                     }
                     if (strcmp(noarg, argdef[i].arg_name) == 0) {
                        bestmatch = i;
                        negate = true;
                        break;
                     } else if (_stricmp(noarg, argdef[i].arg_name) == 0) {
                        bestmatch = i;
                        negate = true;
                     }
                  }
               }
            }
         }
         if (bestmatch == -1) {
            FTHROW(new SimpleException("Unrecognized argument: %s", curarg));
         } else {
            i = bestmatch;
            int arglen = strlen(argdef[i].arg_name);
            const char * argval = NULL;
            if ((! negate) &&
                ((curarg[arglen] == '=') || (curarg[arglen] == ':'))) {
               argval = curarg + arglen + 1;
            }
            if ((argval == NULL) && (argdef[i].arg_kind != WINCMD_BOOL)) {
               if (skip >= argv.size()) {
                  FTHROW(new SimpleException(
                                "Missing argument for option: %s", curarg));
               }
               argval = argv[skip].c_str();
               ++skip;
            }
            switch(argdef[i].arg_kind) {
            case WINCMD_STRING:
               *(const char **)(argdef[i].arg_val_ptr) = argval;
               argdef[i].arg_origin = WINCMD_COMMAND_LINE;
               break;
            case WINCMD_LONG:
               *(long *)(argdef[i].arg_val_ptr) = atol(argval);
               argdef[i].arg_origin = WINCMD_COMMAND_LINE;
               break;
            case WINCMD_BOOL:
               if (argval == NULL) {
                  *(bool *)(argdef[i].arg_val_ptr) = ! negate;
                  argdef[i].arg_origin = WINCMD_COMMAND_LINE;
               } else {

                  // Allow boolean arguments to be true if the arg
                  // value given matches first character of true,
                  // yes, affirmative, or 1
                  //
                  // Allow them to be negated if match is for
                  // false, no, negative, or 0

                  switch(*argval) {
                  case 't':
                  case 'T':
                  case 'y':
                  case 'Y':
                  case 'a':
                  case 'A':
                  case '1':
                     *(bool *)(argdef[i].arg_val_ptr) = true;
                     argdef[i].arg_origin = WINCMD_COMMAND_LINE;
                     break;
                  case 'f':
                  case 'F':
                  case 'n':
                  case 'N':
                  case '0':
                     *(bool *)(argdef[i].arg_val_ptr) = false;
                     argdef[i].arg_origin = WINCMD_COMMAND_LINE;
                     break;
                  default:
                     FTHROW(new SimpleException(
                                   "Unrecognized value for option: %s",
                                   curarg));
                  }
               }
               break;
            }
         }
      } else {
         break;
      }
   }

   // Now go through arguments again. If we didn't see something on the
   // command line, see if we can find it in the registry instead.  (If we
   // don't find an exact match for the kind of registry entry we want, we
   // don't call it an error, we just ignore the entry).

   for (i = 0; i < argdef_count; ++i) {
      if ((argdef[i].arg_origin == WINCMD_UNSPECIFIED) &&
          (argdef[i].arg_which_registry != NULL) &&
          (argdef[i].arg_registry_key != NULL) &&
          (argdef[i].arg_registry_value != NULL)) {

         // Didn't see this one on command line, see if it is in
         // the registry...

         HKEY optkey;
         if (RegOpenKeyEx(argdef[i].arg_which_registry,
                          argdef[i].arg_registry_key,
                          0,
                          KEY_READ,
                          &optkey) == ERROR_SUCCESS) {
            DWORD need_space = 0;
            DWORD valkind = 0;
            char dummy;
            if ((RegQueryValueEx(optkey,
                                 argdef[i].arg_registry_value,
                                 0,
                                 &valkind,
                                 (LPBYTE)&dummy,
                                 &need_space) == ERROR_MORE_DATA) &&
                (valkind == argdef[i].arg_value_type)) {
               if (valkind == REG_DWORD) {
                  DWORD val;
                  if (RegQueryValueEx(optkey,
                                      argdef[i].arg_registry_value,
                                      0,
                                      &valkind,
                                      (LPBYTE)&val,
                                      &need_space) == ERROR_SUCCESS) {
                     if (argdef[i].arg_kind == WINCMD_LONG) {
                        *(long *)(argdef[i].arg_val_ptr) = (long) val;
                        argdef[i].arg_origin = WINCMD_REGISTRY;
                     } else if (argdef[i].arg_kind == WINCMD_BOOL) {
                        *(bool *)(argdef[i].arg_val_ptr) = (bool) (val != 0);
                        argdef[i].arg_origin = WINCMD_REGISTRY;
                     }
                  }
               } else if (valkind == REG_SZ) {
                  char * buf = new char [need_space + 1];
                  if (RegQueryValueEx(optkey,
                                      argdef[i].arg_registry_value,
                                      0,
                                      &valkind,
                                      (LPBYTE)buf,
                                      &need_space) == ERROR_SUCCESS) {
                     if (argdef[i].arg_kind == WINCMD_STRING) {
                        *(char **)(argdef[i].arg_val_ptr) = buf;
                        argdef[i].arg_origin = WINCMD_REGISTRY;
                     }
                  }
               }
            }
            RegCloseKey(optkey);
         }
      }
   }
   FTEND
}

void
Wincmd::AppendHelpInfo(std::string & msg, ArgDef * argdef, int argdef_count) {
   FTBEG(("Wincmd::AppendHelpInfo()"))
   for (int i = 0; i < argdef_count; ++i) {
      msg.append(1, '-');
      msg.append(argdef[i].arg_name);
      switch(argdef[i].arg_kind) {
      case WINCMD_STRING:
         msg.append("=string ");
         break;
      case WINCMD_LONG:
         msg.append("=int ");
         break;
      case WINCMD_BOOL:
         msg.append(" (flag) ");
         break;
      }
      msg.append(argdef[i].arg_description);
      msg.append(1, '\n');
   }
   FTEND
}
